from torch.utils.data import dataset

from ..environment import np, pd, plt,torch,torchvision

"""  
    1.构建一个 PyTorch 数据迭代器。

    Parameters:
    - data_arrays (tuple of torch.Tensor): 包含输入数据和标签数据的元组。
    - batch_size (int): 每个批次的样本数量。
    - is_train (bool, optional): 指定是否为训练模式，默认为 True。

    Returns:
    - torch.utils.data.DataLoader: PyTorch 数据加载器对象，用于迭代数据批次。

    Defined in :numref:`sec_utils`
"""

def load_arrays(data_arrays, batch_size, is_train=True):
    # 创建一个 PyTorch 数据集对象，其中包含输入数据和标签数据
    dataset = torch.utils.data.TensorDataset(*data_arrays)

    # 创建一个 PyTorch 数据加载器对象，用于批次化地加载数据
    # 参数 shuffle 指定是否在每个 epoch 开始前打乱数据顺序
    return torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=is_train)




from torch.utils.data import dataset

from ..environment import np, pd, plt,torch,torchvision





"""  
   1.加载MNIST数据集  

   参数:  
   batch_size (int): 每个batch的大小  
   train (bool): 如果为True，则加载训练集；如果为False，则加载测试集  
   download (bool): 如果为True，则从互联网上下载数据集（如果尚未下载）  

   返回:  
   dataloader (torch.utils.data.DataLoader): 数据加载器  
   
    # 加载训练集  
    train_loader = load_data_mnist(batch_size=64, train=True)  
  
    # 加载测试集  
    test_loader = load_data_mnist(batch_size=64, train=False)
   
"""
def load_data_minist(batch_size=256, resize=None, train=True, download=True):
    # 定义数据转换
    trans = [torchvision.transforms.ToTensor()]
    if resize:
        trans.insert(0,torchvision.transforms.Resize(resize))
    trans = torchvision.transforms.Compose(trans)

    # 根据train参数选择数据集
    if train:
        mnist_dataset = torchvision.datasets.MNIST(root='./data', train=True, transform=trans, download=download)
    else:
        mnist_dataset = torchvision.datasets.MNIST(root='./data', train=False, transform=trans, download=download)

    # 创建数据加载器
    dataloader = torch.utils.data.DataLoader(mnist_dataset, batch_size=batch_size, shuffle=True)

    return dataloader











"""  
    2.加载Fashion-MNIST数据集  

    参数:  
    batch_size (int): 每个batch的大小  
    train (bool): 如果为True，则加载训练集；如果为False，则加载测试集  
    download (bool): 如果为True，则从互联网上下载数据集（如果尚未下载）  

    返回:  
    dataloader (torch.utils.data.DataLoader): 数据加载器  
    
    
    # 加载训练集  
    train_loader = load_data_fashion_mnist(batch_size=64, train=True)  
  
    # 加载测试集  
    test_loader = load_data_fashion_mnist(batch_size=64, train=False)
"""
def load_data_fashion_mnist(batch_size=256, resize=None, train=True, download=True):
    # 定义数据转换
    trans = [torchvision.transforms.ToTensor()]
    if resize:
        trans.insert(0,torchvision.transforms.Resize(resize))
    trans = torchvision.transforms.Compose(trans)

    # 根据train参数选择数据集
    if train:
        fashion_mnist_dataset = torchvision.datasets.FashionMNIST(root='./data', train=True, transform=trans, download=download)
    else:
        fashion_mnist_dataset = torchvision.datasets.FashionMNIST(root='./data', train=False, transform=trans, download=download)

    # 创建数据加载器
    dataloader = torch.utils.data.DataLoader(fashion_mnist_dataset, batch_size=batch_size, shuffle=True)

    return dataloader









import numpy as np

from ..environment import torch, cv2, plt




'''
1.visualize_grid_cells(image_path, W, H, X, Y)
将图像划分的grid cell进行可视化
image_path：图像的路径
W:设置图像的宽
H：设置图像的高
X、Y：设置grid cell的数量为X行×Y列个
'''
def visualize_grid_cells(image_path, W, H, ROW, COLUMN, Title='Resized Image with Grid Cells'):
    # 读取图像
    image = cv2.imread(image_path)

    # 如果图像为空（例如文件不存在或损坏），则退出
    if image is None:
        print(f"Error: Could not read image at {image_path}")
        return

        # 调整图像大小
    resized_image = cv2.resize(image, (W, H))

    # 划分grid cell
    cell_width = W // COLUMN
    cell_height = H // ROW

    # 遍历每个grid cell，绘制边界
    for i in range(COLUMN):
        for j in range(ROW):
            # 计算grid cell的左上角坐标
            start_x = i * cell_width
            start_y = j * cell_height
            # 计算grid cell的右下角坐标
            end_x = min((i + 1) * cell_width, W)  # 防止超出图像宽度
            end_y = min((j + 1) * cell_height, H)  # 防止超出图像高度

            # 绘制grid cell边界（使用BGR颜色，因为在OpenCV中图像是BGR格式的）
            cv2.rectangle(resized_image, (start_x, start_y), (end_x, end_y), (0, 255, 0), 2)

            # 显示原图与grid cell（合并到一张图像）
    plt.imshow(cv2.cvtColor(resized_image, cv2.COLOR_BGR2RGB))
    plt.title(Title)
    plt.axis('off')
    plt.show()





'''
2.show_images
展示图像
'''
def show_images(imgs, num_rows, num_cols, titles=None, scale=1.5, normalize = False):
    # 绘制图像列表
    figsize = (num_cols * scale, num_rows * scale)
    _, axes = plt.subplots(num_rows, num_cols, figsize=figsize)
    axes = axes.flatten()
    for i, (ax, img) in enumerate(zip(axes, imgs)):
        if torch.is_tensor(img):
            # 图像张量
            if img.dim() == 3:
                img = img.permute(1, 2, 0)  # 将通道维度移动到最后的位置
            if normalize:
                img = img.numpy().clip(0, 1)  # 归一化到0-1范围，并裁剪超出范围的值
            ax.imshow(img)

        else:
            # PIL图像
            ax.imshow(img)
        ax.axes.get_xaxis().set_visible(False)
        ax.axes.get_yaxis().set_visible(False)
        if titles:
            ax.set_title(titles[i])
    return axes





'''
3.box_corner_to_center(boxes)
    summary: it transform the box corners to center
'''
def box_corner_to_center(boxes):
    x1, y1, x2, y2 = boxes[:, 0], boxes[:, 1], boxes[:, 2], boxes[:, 3]
    cx = (x1 + x2) // 2
    cy = (y1 + y2) // 2
    w = x2 - x1
    h = y2 - y1
    boxes = torch.stack([cx, cy, w, h], axis=-1)
    return boxes





'''
4.bbox_to_rect
    summary: it draw bounding box rectangle
'''
def bbox_to_rect(bbox, color):
    return plt.Rectangle(xy=(bbox[0], bbox[1]),width=bbox[2]-bbox[0],height=bbox[3]-bbox[1],fill=False,edgecolor=color,linewidth=2)


'''
5.show_boxes(axes, bboxes, labels=None, colors=None)
    summary: it show the box corners
    
'''
def show_boxes(axes, bboxes, labels, colors=None):
    """显示所有的边框"""
    def _make_list(obj, default_values=None):
        if obj is None:
            obj = default_values
        elif not isinstance(obj, (list, tuple)):
            obj = [obj]
        return obj

    labels = _make_list(labels)
    colors = _make_list(colors,['b', 'g', 'r', 'm', 'c' ])
    for i, bbox in enumerate(bboxes):
        color = colors[i % len(colors)]
        rect = bbox_to_rect(bbox.detach().numpy(), color)
        axes.add_patch(rect)
        if labels and len(labels)>i:
            text_color = 'k' if color== 'w' else 'w'
            axes.text(rect.xy[0],rect.xy[1],labels[i],va='center',ha='center',color=text_color,fontsize=9,bbox=dict(facecolor=color, lw=0))


# 5.visualize_bounding_boxes
def visualize_bounding_boxes(image, bounding_boxes, labels):
    """
    Visualizes bounding boxes on an image.

    Parameters:
        image (numpy.ndarray): The input image.
        bounding_boxes (list): A list of bounding boxes in the format [xmin, ymin, xmax, ymax].
        labels (list): A list of labels corresponding to each bounding box.
    """
    # Create figure and axes
    fig, ax = plt.subplots()

    # Display the image
    ax.imshow(image)

    # Add bounding boxes
    for bbox, label in zip(bounding_boxes, labels):
        xmin, ymin, xmax, ymax = bbox
        # Convert relative coordinates to absolute coordinates
        xmin_abs = xmin * image.shape[1]
        ymin_abs = ymin * image.shape[0]
        xmax_abs = xmax * image.shape[1]
        ymax_abs = ymax * image.shape[0]
        width = xmax_abs - xmin_abs
        height = ymax_abs - ymin_abs
        # Create a Rectangle patch
        rect = plt.Rectangle((xmin_abs, ymin_abs), width, height, fill=False, edgecolor='red', linewidth=2)
        ax.add_patch(rect)
        # Add label
        ax.text(xmin_abs, ymin_abs, label, color='red', fontsize=12, verticalalignment='top')

    plt.show()



'''
6.box_iou_xyxy()
    summary: 计算预测框与真实框之间的IOU值，坐标格式为xyxy
    Args:
        boxes (list): A list of bounding boxes in the format [xmin, ymin, xmax, ymax].
'''
def box_iou_xyxy(boxe1, boxe2):
    # 获取box1的具体坐标
    x1min, y1min, x2max, y2max = boxe1[0], boxe1[1], boxe1[2], boxe1[3]

    # 计算boxe1的框面积
    S1 = np.maximum((x2max - x1min) * (y2max - y1min), 0)

    # 获取box2的具体坐标
    x2min, y2min, x2max, y2max = boxe2[0], boxe2[1], boxe2[2], boxe2[3]

    # 计算boxe2的框面积.
    S2 = np.maximum((x2max - x2min) * (y2max - y2min), 0)

    # 计算相交矩形框的坐标，即boxe1与boxe2的交集
    xmin = np.maximum(x1min, x2min)
    ymin = np.maximum(y1min, y2min)
    xmax = np.minimum(x2max, x2max)
    ymax = np.minimum(y2max, y2max)

    # 计算两个矩形框的交集
    intersection = np.maximum((xmax - xmin) * (ymax - ymin), 0)

    IoU = intersection / np.maximum((S1 + S2 - intersection),1.)

    return IoU



'''
7.box_iou_xywh()
'''
def box_iou_xywh(boxe1, boxe2):
    # 将xywh化为xyxy
    x1min, y1min, x1max, y1max = boxe1[0] - boxe1[2]/2.0 , boxe1[1] - boxe1[3]/2.0, boxe1[0] + boxe1[2]/2.0 , boxe1[1] + boxe1[3]/2.0
    x2min, y2min, x2max, y2max = boxe2[0] - boxe2[2]/2.0 , boxe2[1] - boxe2[3]/2.0, boxe2[0] + boxe2[2]/2.0 , boxe2[1] + boxe2[3]/2.0

    return box_iou_xyxy((x1min, y1min, x1max, y1max), (x2min, y2min, x2max, y2max))




'''
8.draw_anchor_boxes
'''
def draw_anchor_boxes(image, center_x, center_y, widths, heights):
    """
    在图像中画出给定中心点和多种宽高的锚框。

    参数：
        image: 输入的图像，可以是 numpy 数组。
        center_x: 中心点的 x 坐标。
        center_y: 中心点的 y 坐标。
        widths: 锚框的宽度列表。
        heights: 锚框的高度列表。

    返回值：
        annotated_image: 画有锚框的图像，numpy 数组格式。
    """
    # 创建一个副本以保留原始图像
    annotated_image = image.copy()

    # 循环绘制每个锚框
    for width, height in zip(widths, heights):
        # 计算左上角和右下角坐标
        top_left = (int(center_x - width / 2), int(center_y - height / 2))
        bottom_right = (int(center_x + width / 2), int(center_y + height / 2))

        # 在图像上画出锚框
        annotated_image = cv2.rectangle(annotated_image, top_left, bottom_right, (0, 255, 0), 2)

    return annotated_image









'''
9.get_objectness_label(img, gt_boxes, gtlabels, iou_threshold=0.7, anchors = [])
    summary: 标注预测框的objectness
    img 是输入的图像数据，形状是[N, C, H, W]
    gt_boxes，真实框，维度是[N, 50, 4]，其中50是真实框数目的上限，当图片中真实框不足50个时，不足部分的坐标全为0
              真实框坐标格式是xywh，这里使用相对值
    gt_labels，真实框所属类别，维度是[N, 50]
    iou_threshold，当预测框与真实框的iou大于iou_threshold时不将其看作是负样本
    anchors，锚框可选的尺寸
    anchor_masks，通过与anchors一起确定本层级的特征图应该选用多大尺寸的锚框
    num_classes，类别数目
    downsample，特征图相对于输入网络的图片尺寸变化的比例
'''
def get_objectness_label(img, gt_boxes, gtlabels, iou_threshold=0.7, anchors = [116, 90, 156, 198, 373, 326], num_classes = 3, downsample = 32):
    img_shape = img.shape
    batch_size = img.shape[0]
    num_anchors = len(anchors) // 2
    input_h = img.shape[2]
    input_w = img.shape[3]

    # 将输入图片划分成num_rows x num_cols个候选区域，每个候选区域的边长是downsample
    # 计算共有多少列候选区域
    num_cols = input_w // downsample
    # 计算共有多少行候选区域
    num_rows = input_h // downsample

    label_objectness = np.zeros([batch_size, num_anchors, num_rows, num_cols])
    label_classification = np.zeros([batch_size, num_anchors, num_classes, num_rows, num_cols])
    label_location = np.zeros([batch_size, num_anchors, 4, num_rows, num_cols])

    # 这里有个scale_location，用来调节大目标和小目标的权重的
    # 大目标和小目标对于相同的偏移，大目标更加敏感，故增强这个权重系数
    scale_location = np.zeros([batch_size, num_anchors, num_rows, num_cols])

    # batch_size进行循环，处理每一张图片
    for n in range(batch_size):
        # 对图片上的真实框进行遍历，依次找出与真实框形状最匹配的锚框
        for n_gt in range(len(gt_boxes[n])):
            gt_box = gt_boxes[n][n_gt]
            # 真实框所属类别
            gt_label = gtlabels[n][n_gt]
            # 真实框的中心坐标和长宽（YOLO中采用xywh）
            gt_center_x = gt_box[0]
            gt_center_y = gt_box[1]
            gt_width = gt_box[2]
            gt_height = gt_box[3]
            if (gt_width < 1e-3) or (gt_height < 1e-3):
                continue

            # 计算这个真实框的中心点落在哪个格子里,xy是真实边界框的中心点在图像中的归一化坐标（即范围在 [0, 1] 内）
            i = int(gt_center_y * num_rows)
            j = int(gt_center_x * num_cols)

            ious = []

            for ka in range(num_anchors):
                # x，y为0，假设锚框和真实框的x，y相同
                gt_bbox = [0., 0., float(gt_width), float(gt_height)]
                # 生成输入（i，j）这个格子里的锚框
                anchor_w = anchors[ka * 2]
                anchor_h = anchors[ka * 2 + 1]
                anchor_bbox = [0., 0., anchor_w / float(input_w), anchor_h / float(input_h)]
                # 计算iou
                iou = box_iou_xywh(gt_bbox, anchor_bbox)
                ious.append(iou)
            ious = np.array(ious)
            inds = np.argsort(ious)
            k = inds[-1]
            label_objectness[n, k, i, j] = 1
            c = gt_label
            # 使用one-hot方法标注
            label_classification[n, k, c, i, j] = 1

            # for those prediction bbox with objectness =1, set label of location
            # 反算出真实框的tx*，ty*，tw*，th*
            dx_label = gt_center_x * num_cols - j
            dy_label = gt_center_y * num_rows - i
            dw_label = np.log(gt_width * input_w / anchors[k * 2])
            dh_label = np.log(gt_height * input_h / anchors[k * 2 + 1])
            label_location[n, k, 0, i, j] = dx_label
            label_location[n, k, 1, i, j] = dy_label
            label_location[n, k, 2, i, j] = dw_label
            label_location[n, k, 3, i, j] = dh_label

            # scale_location用来调节不同尺寸的锚框对损失函数的贡献，作为加权系数和位置损失函数相乘
            # 根据这个计算方法，可以得出真实框越大，权重越小
            scale_location[n, k, i, j] = 2.0 - gt_width * gt_height

    # 目前根据每张图片上所有出现过的gt_box都标注出了objectness为正的预测框，剩下的预测框的objectness为0
    # 对于objectness为1的预测框，标出它们所包含的物体类别以及位置回归的目标
    return label_objectness.astype('float32'), label_location.astype('float32'), label_classification.astype('float32'), scale_location.astype('float32')






'''
10.NMS非极大值抑制
    summary: Non-Maximum Suppression
'''
def nms(bboxes, scores, score_thresh, nms_thresh, pre_nms_topk, i=0, c=0):
    inds = np.argsort(scores)
    inds = inds[::-1]
    keep_inds = []
    while(len(inds) > 0):
        cur_ind = inds[0]
        cur_score = scores[cur_ind]
        # 如果该框的分数小于score.thresh，则丢弃该框
        if cur_score < score_thresh:
            break

        keep = True
        for ind in keep_inds:
            current_box = bboxes[cur_ind]
            remain_box = bboxes[ind]
            iou = box_iou_xywh(current_box, remain_box)
            if iou > nms_thresh:
                keep = False
                break
        # 调试代码，用于观察信息使用
        if i == 0 and c == 4 and cur_ind == 951:
            print('suppressed, ',keep,i,c,cur_ind,ind,iou)
        if keep:
            keep_inds.append(cur_ind)
        # 去除当前处理的边界框索引，即将 inds 数组中的第一个元素移除，继续处理下一个边界框。
        inds = inds[1:]
    return np.array(keep_inds)






'''
11.multiclass_nms
    summary: multiclass-Non-Maximum Suppression
'''
def multiclass_nms(bboxes, scores, score_thresh = 0.01, nms_thresh = 0.45, pre_nms_topk = 1000, pos_nms_topk = 100 ):
    # 获取批量大小和类别数量
    batch_size = bboxes.shape[0]
    class_num = scores.shape[1]
    rets = []
    # 遍历每个样本
    for i in range(batch_size):
        # 获取当前样本的边界框和分数
        bboxes_i = bboxes[i]
        scores_i = scores[i]
        ret = []
        # 遍历每个类别
        for c in range(class_num):
            # 获取当前类别的分数
            scores_i_c = scores_i[c]
            # 使用 NMS 函数进行非最大抑制
            keep_inds = nms(bboxes_i, scores_i_c, score_thresh, nms_thresh, pre_nms_topk, i, c)

            # 如果保留的边界框数量小于1，则跳过当前类别
            if len(keep_inds) < 1:
                continue
            # 根据保留的索引获取对应的边界框和分数
            keep_bboxes = bboxes_i[keep_inds]
            keep_scores = scores_i_c[keep_inds]
            # 构造保留结果
            keep_results = np.zeros([keep_scores.shape[0], 6])
            keep_results[:, 0] = c # 类别索引
            keep_results[:, 1] = keep_scores[:] # 分数
            keep_results[:, 2:6] = keep_bboxes[:, :]    # 边界框坐标
            ret.append(keep_results)
        # 如果当前样本没有保留的结果，则跳过
        if len(ret) < 1:
            rets.append(ret)
            continue
        # 将每个类别的结果拼接在一起
        ret_i = np.concatenate(ret, axis=0)
        # 如果保留的结果数量超过了指定的阈值，则根据分数进行排序并截取指定数量的结果
        scores_i = ret_i[:, 1]
        if len(scores_i) > pos_nms_topk:
            inds = np.argsort(scores_i)[::-1]
            inds = inds[:pos_nms_topk]
            ret_i = ret_i[inds]

        rets.append(ret_i)

    return rets





























































# layers 模块
# 功能：包含自定义的神经网络层或模块。
# 子模块/文件：
# attention.py：实现各种注意力机制层。
# custom_layer.py：用户自定义层的示例或模板。












# # layers/__init__.py
#
# # 导入 layers 子包中的各个模块
# from . import conv_layers
# from . import pooling_layers
# from . import normalization_layers
# from . import activation_layers
#
# # 如果需要，可以直接导入这些模块中的特定层类
# from .conv_layers import Conv2D, DepthwiseConv2D
# from .pooling_layers import MaxPooling2D, AveragePooling2D
# from .normalization_layers import BatchNormalization
# from .activation_layers import ReLU, Sigmoid, Softmax
#
# # __all__ 变量定义了当使用 from layers import * 时导入哪些对象
# # 注意：通常不推荐使用 from package import *
# __all__ = [
#     'Conv2D', 'DepthwiseConv2D',
#     'MaxPooling2D', 'AveragePooling2D',
#     'BatchNormalization',
#     'ReLU', 'Sigmoid', 'Softmax',
#     # ... 其他希望用户直接访问的关键层类 ...
# ]


from ..environment import torch
from ..environment import collections
from ..environment import random

"""
1.tokenize(lines, token='word')
简述：
tokenize函数将文本行列表（lines）作为输入， 列表中的每个元素是一个文本序列（如一条文本行）。
每个文本序列又被拆分成一个词元列表，词元（token）是文本的基本单位。
最后，返回一个由词元列表组成的列表，其中的每个词元都是一个字符串（string）。
"""
def tokenize(lines, token='word'):
    """将文本行拆分为单词或字符词元"""
    if token == 'word':
        return [line.split() for line in lines]
    elif token == 'char':
        return [list(line) for line in lines]
    else:
        print('错误：未知词元类型：' + token)





"""
2.Vocab
文本词表(vocabulary)
初始化词表对象。

Parameters:
        tokens (list): 用于构建词表的词元列表，默认为 None。
        min_freq (int): 词频阈值，低于该频率的词元将被过滤，默认为 0。
        reserved_tokens (list): 预留的特殊词元列表，默认为 None。

Attributes:
        _token_freqs (list): 按词频降序排列的词元及其频率的列表。
        idx_to_token (list): 索引到词元的映射列表。
        token_to_idx (dict): 词元到索引的映射字典。
"""
class Vocab:

    def __init__(self, tokens=None, min_freq=0, reserved_tokens=None):
        if tokens is None:
            tokens = []
        if reserved_tokens is None:
            reserved_tokens = []


        # 统计词元的频率
        counter = self.count_corpus(tokens)

        # 按出现频率排序
        self._token_freqs = sorted(counter.items(), key=lambda x: x[1], reverse=True)

        # 初始化索引到词元的映射列表，加入预留词元 "<unk>"
        self.idx_to_token = ['<unk>'] + reserved_tokens

        # 初始化词元到索引的映射字典
        self.token_to_idx = {token: idx for idx, token in enumerate(self.idx_to_token)}

        # 将词频大于 min_freq 的词元添加到词表中
        for token, freq in self._token_freqs:
            if freq < min_freq:
                break
            if token not in self.token_to_idx:
                self.idx_to_token.append(token)
                self.token_to_idx[token] = len(self.idx_to_token) - 1

    def __len__(self):
        """返回词表中的词元数量"""
        return len(self.idx_to_token)

    def __getitem__(self, tokens):
        """
        返回给定词元的索引。

        Parameters:
            tokens (str or list): 单个词元或词元列表。

        Returns:
            idx (int or list): 词元对应的索引或索引列表。
        """
        # 如果 tokens 是单个词元，返回其索引；如果是词元列表，则返回对应的索引列表
        if not isinstance(tokens, (list, tuple)):
            return self.token_to_idx.get(tokens, self.unk)
        return [self.__getitem__(token) for token in tokens]

    def to_tokens(self, indices):
        """
        返回给定索引对应的词元。

        Parameters:
            indices (int or list): 单个索引或索引列表。

        Returns:
            tokens (str or list): 索引对应的词元或词元列表。
        """
        # 如果 indices 是单个索引，返回对应的词元；如果是索引列表，则返回对应的词元列表
        if not isinstance(indices, (list, tuple)):
            return self.idx_to_token[indices]
        return [self.idx_to_token[index] for index in indices]

    @property
    def unk(self):
        """返回未知词元的索引"""
        return 0

    @property
    def token_freqs(self):
        """返回词元及其频率的列表"""
        return self._token_freqs

    def count_corpus(self, tokens):  # 改为实例方法
        """
        统计词元的频率。

        Parameters:
            tokens (list): 用于统计的词元列表，可以是一维或二维列表。

        Returns:
            counter (collections.Counter): 词元及其频率的计数器。
        """
        # 如果 tokens 是二维列表，将其展平成一维列表
        if len(tokens) == 0 or isinstance(tokens[0], list):
            tokens = [token for line in tokens for token in line]

        # 统计词元的频率并返回计数器
        return collections.Counter(tokens)



"""  
3.seq_data_iter_random(corpus, batch_size, num_steps)
    使用随机抽样生成一个小批量子序列  

    参数:  
        corpus (list): 文本语料库，一个字符列表  
        batch_size (int): 每个小批量的样本数量  
        num_steps (int): 每个子序列的长度  

    返回:  
        生成器: 产生包含输入X和标签Y的小批量数据  
"""
def seq_data_iter_random(corpus, batch_size, num_steps):
    """使用随机抽样生成一个小批量子序列"""
    # 从随机偏移量开始对序列进行分区，随机范围包括num_steps-1
    corpus = corpus[random.randint(0, num_steps - 1):]

    # 减去1，是因为我们需要考虑标签
    num_subseqs = (len(corpus) - 1) // num_steps

    # 长度为num_steps的子序列的起始索引
    initial_indices = list(range(0, num_subseqs * num_steps, num_steps))

    # 在随机抽样的迭代过程中，
    # 来自两个相邻的、随机的、小批量中的子序列不一定在原始序列上相邻
    random.shuffle(initial_indices)


    def data(pos):
        # 返回从pos位置开始的长度为num_steps的序列
        return corpus[pos: pos + num_steps]


    num_batches = num_subseqs // batch_size
    for i in range(0, batch_size * num_batches, batch_size):
        # 在这里，initial_indices包含子序列的随机起始索引
        initial_indices_per_batch = initial_indices[i: i + batch_size]
        X = [data(j) for j in initial_indices_per_batch]
        Y = [data(j + 1) for j in initial_indices_per_batch]
        yield torch.tensor(X), torch.tensor(Y)














"""  
4.seq_data_iter_random(corpus, batch_size, num_steps)
    顺序抽样生成一个小批量子序列  

    参数:  
        corpus (list): 文本语料库，一个字符列表  
        batch_size (int): 每个小批量的样本数量  
        num_steps (int): 每个子序列的长度  

    返回:  
        生成器: 产生包含输入X和标签Y的小批量数据  
"""
def seq_data_iter_sequential(corpus, batch_size, num_steps):
    """使用顺序分区生成一个小批量子序列"""
    # 从随机偏移量开始划分序列
    offset = random.randint(0, num_steps)
    num_tokens = ((len(corpus) - offset - 1) // batch_size) * batch_size
    Xs = torch.tensor(corpus[offset: offset + num_tokens])
    Ys = torch.tensor(corpus[offset + 1: offset + 1 + num_tokens])
    Xs, Ys = Xs.reshape(batch_size, -1), Ys.reshape(batch_size, -1)
    num_batches = Xs.shape[1] // num_steps
    for i in range(0, num_steps * num_batches, num_steps):
        X = Xs[:, i: i + num_steps]
        Y = Ys[:, i: i + num_steps]
        yield X, Y
















# 5.SeqDataLoader
class SeqDataLoader:  # @save
    """
    加载序列数据的迭代器。
    这个类提供了两种加载序列数据的方式：随机抽样（seq_data_iter_random）和顺序抽取（seq_data_iter_sequential）。
    """

    def __init__(self, lines, token,batch_size, num_steps, use_random_iter, max_tokens):
        """
        初始化序列数据加载器。

        参数:
        - batch_size (int): 每个小批量的样本数量。
        - num_steps (int): 每个子序列的长度。
        - use_random_iter (bool): 是否使用随机抽样来生成子序列。
        - max_tokens (int): 加载语料库时使用的最大令牌数量。

        属性:
        - data_iter_fn (callable): 根据use_random_iter的值，选择随机或顺序的迭代函数。
        - corpus (list): 文本语料库，一个字符列表。
        - vocab (d2l.Vocab): 词汇表对象，用于将字符映射到索引。
        - batch_size (int): 每个小批量的样本数量。
        - num_steps (int): 每个子序列的长度。
        """
        if use_random_iter:
            # 如果use_random_iter为True，则使用随机抽样的迭代函数
            self.data_iter_fn = seq_data_iter_random
        else:
            # 否则，使用顺序抽取的迭代函数
            self.data_iter_fn = seq_data_iter_sequential

            # 加载语料库并限制最大令牌数量，同时获取词汇表对象
        tokens = tokenize(lines, token)
        self.vocab = Vocab(tokens)
        self.corpus = [self.vocab[token] for line in tokens for token in line]
        if max_tokens>0:
            self.corpus = self.corpus[:max_tokens]



        # 保存批大小和子序列长度
        self.batch_size, self.num_steps = batch_size, num_steps

    def __iter__(self):
        """
        实现迭代器协议，返回一个迭代器，该迭代器生成小批量数据。

        返回:
        - generator: 一个生成器，产生包含输入X和标签Y的小批量数据。
        """
        # 使用选定的迭代函数（随机或顺序）生成数据
        return self.data_iter_fn(self.corpus, self.batch_size, self.num_steps)




































from ..environment import torch, random, np, cv2

'''
1.Random_Expand
    summary: the Random_Expand class intend to implement the random padding method for image
    If the image is torch Tensor, it is expected to have [..., H, W] shape.

    Args:
        max_ratio: Maximum ratio for random expansion. Default is 4.
        fill: Fill color for the expanded area. Default is None.
        keep_ratio: Wether to maintain aspect ratio during expansion. Default is True.
        thresh: Threshold for applying the random expansion. Default is 0.5.

'''
class Random_Expand(torch.nn.Module):
    def __init__(self, max_ratio = 4, fill = None, keep_ratio = True, thresh = 0.5):
        super(Random_Expand, self).__init__()
        if not isinstance(max_ratio, (int, float)):
            raise TypeError('max_ratio must be an int or float')

        if not isinstance(fill, (int, tuple, type(None))):
            raise TypeError('fill must be an int or tuple')

        if not isinstance(keep_ratio, bool):
            raise TypeError('keep_ratio must be bool')

        if not isinstance(thresh, (int, float)):
            raise TypeError('thresh must be an int or float')

        self.max_ratio = max_ratio
        self.fill = fill
        self.keep_ratio = keep_ratio
        self.thresh = thresh

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(max_ratio={self.max_ratio}, fill={self.fill}, keep_ratio={self.keep_ratio}, thresh={self.thresh})"

    def forward(self, image, targets=None):
        # Check if random expansion should be applied based on the threshold
        if random.random() > self.thresh:  # If not, return the original image and bounding boxes
            return image, targets

        # Check if the maximum expansion ratio is less than 1 or not properly specified
        if self.max_ratio < 1.0:
            return image, targets  # If so, return the original image and bounding boxes

        # Get the height, width, and number of channels of the input image
        h, w, c = image.shape

        # Generate random expansion ratios for width and height
        ratio_x = random.uniform(1, self.max_ratio)
        if self.keep_ratio:
            ratio_y = ratio_x  # If keeping ratio, use the same ratio for both dimensions
        else:
            ratio_y = random.uniform(1, self.max_ratio)  # Otherwise, generate a separate ratio for height

        # Calculate the new height and width after expansion
        oh = int(h * ratio_y)
        ow = int(w * ratio_x)

        # Generate random offsets for placing the original image within the expanded canvas
        off_x = random.randint(0, ow - w)
        off_y = random.randint(0, oh - h)

        # Create an empty canvas for the expanded image
        out_img = np.zeros((oh, ow, c))

        # Fill the canvas with a specified color if provided
        if self.fill and len(self.fill) == c:
            for i in range(c):
                out_img[:, :, i] = self.fill[i] * 255.0

        # Place the original image onto the expanded canvas at the generated offset
        out_img[off_y:off_y + h, off_x:off_x + w, :] = image

        if targets is not None:
            # Update the bounding box coordinates to reflect the expansion
            targets[:, 0] = ((targets[:, 0] * w) + off_x) / float(ow)  # Adjust xmin-coordinates
            targets[:, 1] = ((targets[:, 1] * h) + off_y) / float(oh)  # Adjust ymin-coordinates
            targets[:, 2] = ((targets[:, 2] * w) + off_x) / float(ow)  # Adjust xmax-coordinates
            targets[:, 3] = ((targets[:, 3] * h) + off_y) / float(oh)  # Adjust ymax-coordinates

        return out_img.astype('uint8') if not targets else (out_img.astype('uint8'), targets)






'''
2.Random_Interp
    summary: 
        the Random_Interp class intend to implement the random interp method for image
        If the image is torch Tensor, it is expected to have [..., H, W] shape.
        
    Args:
        size: Size of the image. Default is 256.
        interp: Interpolation method. Default is None.
'''
class Random_Interp(torch.nn.Module):
    def __init__(self, size=256, interp = None):
        super().__init__()
        if not isinstance(size, (int, float)):
            raise TypeError('size must be an int or float')

        if not isinstance(interp, (type(None), type(cv2.INTER_CUBIC))):
            raise TypeError('interp must be None or cv2.INTER_CUBIC')
        self.size = size
        self.interp = interp
        self.interp_method = [
            cv2.INTER_CUBIC,
            cv2.INTER_NEAREST,
            cv2.INTER_LANCZOS4,
            cv2.INTER_LINEAR,
            cv2.INTER_AREA
        ]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(size={self.size}, interp={self.interp})"

    def forward(self, image, targets=None):
        if not self.interp or self.interp not in self.interp_method:
            interp = self.interp_method[random.randint(0, len(self.interp_method) - 1)]

        h, w, _ = image.shape
        im_scale_x = self.size / float(w)
        im_scale_y = self.size / float(h)
        image = cv2.resize(image, None, None, fx=im_scale_x, fy=im_scale_y, interpolation=interp)

        return image if not targets else (image, targets)




'''
3.Random_Horizontal_Flip
    summary:
        the Random_Horizontal_Flip class intend to implement the random horizontal flip method for image
        If the image is torch Tensor, it is expected to have [..., H, W] shape.
    Args:
        thresh: Threshold for applying the random horizontal flip. Default is 0.5.
'''
class Random_Horizontal_Flip(torch.nn.Module):
    def __init__(self, thresh = 0.5):
        super().__init__()
        if not isinstance(thresh, (int, float)):
            raise TypeError('thresh must be an int or float')
        self.thresh = thresh

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(thresh={self.thresh})"

    def forward(self, image, targets=None):
        if random.random() > self.thresh:
            image = image[:,::-1,:]
            if targets is not None:
                targets[:, 0] = 1-targets[:, 0]
                targets[:, 2] = 1-targets[:, 2]

        return image if not targets else (image, targets)




'''
4.Random_Shuffle_targets
    summary: Random_Shuffle_targets intend to implement the random shuffle method for order of targets
'''
class Random_Shuffle_targets(torch.nn.Module):
    def __init__(self):
        super().__init__()
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"
    def forward(self, targets):
        idx = np.arange(targets[0])
        np.random.shuffle(idx)
        targets = targets[idx, :]
        return targets



# datasets 模块
# 功能：包含数据集的加载、预处理和增强功能。
# 子模块/文件：
# image_datasets.py：图像数据集的处理，如CIFAR、ImageNet等。
# text_datasets.py：文本数据集的处理，如IMDB、WikiText等。
# custom_dataset.py：用户自定义数据集的示例或模板。










# datasets/__init__.py

# 导入 datasets 子包中的各个模块
# from . import mnist_dataset
# from . import cifar10_dataset
# from . import custom_dataset




# 如果需要，可以直接导入这些模块中的特定类或函数
from .image_datasets import load_data_minist
from .image_datasets import load_data_fashion_mnist


from .custom_dataset import load_arrays
# from .cifar10_dataset import load_cifar10_data
# from .custom_dataset import load_custom_data
#
# __all__ 变量定义了当使用 from datasets import * 时导入哪些对象
# 注意：通常不推荐使用 from package import *
__all__ = [
    'load_data_minist',
    'load_data_fashion_mnist',
    'load_arrays'
]






"""
core 模块
功能：包含包的核心功能和基础类。
子模块/文件：
base_model.py：定义基础模型类，提供模型初始化、保存、加载等基础功能。
utils.py：包含一些通用的工具函数，如数据处理、模型评估等。
text.utils: 负责处理文本数据相关的功能，例如词表构建、词频统计等。
"""
# core/__init__.py

# 导入 core 子包中的关键模块
# from . import data_structures
# from . import utils
# from . import model_builder
#
# 如果需要，可以直接导入这些模块中的特定类或函数

from .utils import evaluate_accuracy_gpu
from .utils import train_custom
from .utils import try_gpu
from .utils import try_all_gpus

from .text_utils import tokenize
from .text_utils import Vocab
from .text_utils import seq_data_iter_random
from .text_utils import seq_data_iter_sequential
from .text_utils import SeqDataLoader

from .image_utils import visualize_grid_cells
from .image_utils import show_images
from .image_utils import show_boxes
from .image_utils import bbox_to_rect
from .image_utils import box_corner_to_center
from .image_utils import visualize_bounding_boxes
from .image_utils import box_iou_xyxy
from .image_utils import box_iou_xywh
from .image_utils import draw_anchor_boxes
from .image_utils import get_objectness_label
from .image_utils import nms
from .image_utils import multiclass_nms

from . import transforms




# from .data_structures import CustomDataStructure
# from .utils import some_utility_function
# from .model_builder import build_model
#
# __all__ 变量定义了当使用 from core import * 时导入哪些对象
# 注意：通常不推荐使用 from package import *
__all__ = [
    'evaluate_accuracy_gpu',
    'train_custom',
    'try_gpu',
    'try_all_gpus',

    'tokenize',
    'Vocab',
    'seq_data_iter_random',
    'seq_data_iter_sequential',
    'SeqDataLoader',

    'visualize_grid_cells',
    'show_images',
    'show_boxes',
    'bbox_to_rect',
    'box_corner_to_center',
    'visualize_bounding_boxes',
    'box_iou_xyxy',
    'box_iou_xywh',
    'draw_anchor_boxes',
    'get_objectness_label',
    'nms',
    'multiclass_nms',

    'transforms'
]


# losses 模块
# 功能：包含损失函数的实现或扩展。
# 子模块/文件：
# custom_loss.py：用户自定义损失函数的示例或模板。












# # losses/__init__.py
#
# # 导入 losses 子包中的各个模块
# from . import mean_squared_error
# from . import categorical_crossentropy
# from . import custom_loss
#
# # 如果需要，可以直接导入这些模块中的特定损失函数或类
# from .mean_squared_error import mean_squared_error_loss
# from .categorical_crossentropy import categorical_crossentropy_loss
# from .custom_loss import custom_loss_function
#
# # __all__ 变量定义了当使用 from losses import * 时导入哪些对象
# # 注意：通常不推荐使用 from package import *
# __all__ = [
#     'mean_squared_error_loss',
#     'categorical_crossentropy_loss',
#     'custom_loss_function',
#     # ... 其他希望用户直接访问的关键损失函数或类 ...
# ]


from ..environment import torch



'''
1.RNNModel
简述：循环神经网络模型
'''
class RNNModel(torch.nn.Module):
    def __init__(self, rnn_layer, vocab_size, **kwargs):
        super(RNNModel, self).__init__(**kwargs)
        self.rnn = rnn_layer  # 循环神经网络层，可以是nn.RNN、nn.LSTM或nn.GRU
        self.vocab_size = vocab_size  # 词表大小，用于全连接层输出
        self.num_hiddens = self.rnn.hidden_size  # 隐藏单元数，等于循环神经网络层的隐藏单元数

        # 如果RNN是双向的（之后将介绍），num_directions应该是2，否则应该是1
        if not self.rnn.bidirectional:
            self.num_directions = 1
            # 线性层，用于将RNN的输出映射到词表大小的空间上
            self.linear = torch.nn.Linear(self.num_hiddens, self.vocab_size)
        else:
            self.num_directions = 2
            # 如果是双向的RNN，线性层的输入维度是隐藏单元数的两倍
            self.linear = torch.nn.Linear(self.num_hiddens * 2, self.vocab_size)

    def forward(self, inputs, state):
        # 将输入转换为one-hot编码，并转换为浮点型张量
        X = torch.nn.functional.one_hot(inputs.T.long(), self.vocab_size)
        X = X.to(torch.float32)
        # 前向传播过程，获取RNN的输出和新的隐状态
        Y, state = self.rnn(X, state)
        # 全连接层首先将Y的形状改为(时间步数*批量大小,隐藏单元数)
        # 它的输出形状是(时间步数*批量大小,词表大小)。
        output = self.linear(Y.reshape((-1, Y.shape[-1])))
        return output, state

    def begin_state(self, device, batch_size=1):
        if not isinstance(self.rnn, torch.nn.LSTM):
            # nn.GRU以张量作为隐状态
            return  torch.zeros((self.num_directions * self.rnn.num_layers,
                                 batch_size, self.num_hiddens),
                                device=device)
        else:
            # nn.LSTM以元组作为隐状态
            return (torch.zeros((
                self.num_directions * self.rnn.num_layers,
                batch_size, self.num_hiddens), device=device),
                    torch.zeros((
                        self.num_directions * self.rnn.num_layers,
                        batch_size, self.num_hiddens), device=device))


# custom_deep_learning/__init__.py

# 导出子模块，但不导出子模块中的具体函数或类
from . import core
from . import models
from . import layers
from . import datasets
from . import optimizers
from . import losses
from . import callbacks
from . import examples
from . import tools



__all__ = ['core', 'layers', 'datasets', 'optimizers', 'losses', 'callbacks', 'examples', 'tools']





# 你可以在这里添加一些文档字符串，描述这个包的功能和使用方式
"""  
自定义深度学习包，用于完成深度学习相关的操作和复现代码。  

子模块：  
    core: 包含包的核心功能和基础类。  
    models: 包含各种深度学习模型的实现。  
    layers: 包含自定义的神经网络层或模块。  
    datasets: 包含数据集的加载、预处理和增强功能。  
    optimizers: 包含优化器的实现或扩展。  
    losses: 包含损失函数的实现或扩展。  
    callbacks: 包含训练过程中的回调函数。  
    examples: 包含使用本包进行深度学习任务的示例代码。  
    tools:包含直接自定义工具
"""





"""
当规划一个自定义的深度学习包时，合理的包结构对于提高代码的可读性、可维护性和复用性至关重要。
以下是一个建议的包结构及其模块命名和功能描述：

包名: PepperPepper

模块:

core 模块
功能：包含包的核心功能和基础类。
子模块/文件：
base_model.py：定义基础模型类，提供模型初始化、保存、加载等基础功能。
utils.py：包含一些通用的工具函数，如数据处理、模型评估等。
text.utils: 负责处理文本数据相关的功能，例如词表构建、词频统计等。

models 模块
功能：包含各种深度学习模型的实现。
子模块/文件：
cnn.py：卷积神经网络（CNN）相关模型的实现。
rnn.py：循环神经网络（RNN）相关模型的实现。
transformer.py：Transformer模型及其变体的实现。
custom_model.py：用户自定义模型的示例或模板。

layers 模块
功能：包含自定义的神经网络层或模块。
子模块/文件：
attention.py：实现各种注意力机制层。
custom_layer.py：用户自定义层的示例或模板。

datasets 模块
功能：包含数据集的加载、预处理和增强功能。
子模块/文件：
image_datasets.py：图像数据集的处理，如CIFAR、ImageNet等。
text_datasets.py：文本数据集的处理，如IMDB、WikiText等。
custom_dataset.py：用户自定义数据集的示例或模板。

optimizers 模块
功能：包含优化器的实现或扩展。
子模块/文件：
custom_optimizer.py：用户自定义优化器的示例或模板。

losses 模块
功能：包含损失函数的实现或扩展。
子模块/文件：
custom_loss.py：用户自定义损失函数的示例或模板。

callbacks 模块
功能：包含训练过程中的回调函数，用于实现如学习率调整、模型保存等功能。
子模块/文件：
learning_rate_scheduler.py：学习率调整策略的实现。
custom_callback.py：用户自定义回调函数的示例或模板。

examples 模块
功能：包含使用custom_deep_learning包进行深度学习任务的示例代码。
子模块/文件：
image_classification.py：使用CNN进行图像分类的示例。
text_generation.py：使用Transformer进行文本生成的示例。
custom_task.py：用户自定义任务的示例。
这样的包结构使得每个模块都有其明确的功能和职责，方便用户根据需要进行查找和使用。同时，用户还可以根据具体需求在已有模块的基础上进行扩展或添加新的模块。
"""



"""
__init__.py 文件起到了非常关键的作用，它用于初始化包，并可以定义一些公共的导入或设置。以下是如何设计您的自定义包结构以及编写 __init__.py 文件的一些建议。

包结构设计
首先，根据您之前提供的模块列表，我们可以组织包结构如下：

custom_deep_learning/  
├── __init__.py  
├── core/  
│   ├── __init__.py  
│   ├── base_model.py  
│   └── utils.py  
├── models/  
│   ├── __init__.py  
│   ├── cnn.py  
│   ├── rnn.py  
│   ├── transformer.py  
│   └── custom_model.py  
├── layers/  
│   ├── __init__.py  
│   ├── attention.py  
│   └── custom_layer.py  
├── datasets/  
│   ├── __init__.py  
│   ├── image_datasets.py  
│   ├── text_datasets.py  
│   └── custom_dataset.py  
├── optimizers/  
│   ├── __init__.py  
│   └── custom_optimizer.py  
├── losses/  
│   ├── __init__.py  
│   └── custom_loss.py  
├── callbacks/  
│   ├── __init__.py  
│   ├── learning_rate_scheduler.py  
│   └── custom_callback.py  
└── examples/  
    ├── __init__.py  
    ├── image_classification.py  
    ├── text_generation.py  
    └── custom_task.py
"""





from ..environment import np, pd, plt, torch, torchvision, cv2
from ..callbacks import save_best_model







"""  
1.evaluate_accuracy_gpu(net, data_iter, device=None)
评估模型在给定数据集上的准确性，并将数据加载到指定的GPU设备上（如果可用）。  

Args:  
    net (torch.nn.Module): 需要评估的神经网络模型。  
    data_iter (Iterable): 数据迭代器，提供图像和标签的批量数据。  
    device (torch.device, optional): 要将数据加载到的设备。默认为None，表示如果GPU可用则使用第一个GPU，否则使用CPU。  

Returns:  
    float: 模型的准确性（正确预测的样本数占总样本数的比例）。  

"""
def evaluate_accuracy_gpu(net, data_iter, device=None):
    if device is None and torch.cuda.is_available():
        device = torch.device('cuda')  # 如果未指定设备且GPU可用，则使用GPU
    elif device is None:
        device = torch.device('cpu')  # 如果未指定设备且GPU不可用，则使用CPU


    net.eval()  # 设置为评估模式
    acc_sum = 0
    num_examples = 0

    with torch.no_grad():
        for images, labels in data_iter:
            # 将图像和标签移动到指定的设备
            images, labels = images.to(device), labels.to(device)
            # 前向传播，获取预测结果
            outputs = net(images)
            # 获取预测结果中概率最大的类别作为预测类别
            _, predicted  = torch.max(outputs, 1)
            # 累加样本数量和正确预测的样本数量
            num_examples += labels.size(0)
            acc_sum += (predicted == labels).sum().item()

    # 返回准确率
    return acc_sum / num_examples













""" 
    2.train_custom(model, train_loader, valid_loader, epochs, lr, device, model_path)
    自定义的模型训练函数，包含保存最佳模型的功能  

    参数:  
        model (nn.Module): 待训练的神经网络模型  
        train_loader (DataLoader): 训练数据加载器  
        valid_loader (DataLoader): 验证数据加载器  
        epochs (int): 训练轮数  
        lr (float): 学习率  
        device (torch.device): 设备类型（CPU或GPU）  
        model_path (str): 最佳模型保存路径  

    返回:  
        None  
"""

def train_custom(model, train_loader, valid_loader, epochs, lr, device, model_path):
    # 用于评估模型准确率的函数，同时返回验证损失和准确率
    def evaluate_accuracy(model, valid_loader, criterion, device):
        model.eval()
        val_loss = 0.0
        correct = 0
        total = 0
        with torch.no_grad():
            for images, labels in valid_loader:
                images, labels = images.to(device), labels.to(device)
                outputs = model(images)
                loss = criterion(outputs, labels)
                val_loss += loss.item() * images.size(0)
                _, predicted = torch.max(outputs, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()
        val_acc = 100 * correct / total
        return val_loss / total, val_acc

        # 初始化模型参数，使用xavier方法
    def init_weights(m):
        if isinstance(m, (torch.nn.Linear, torch.nn.Conv2d)):  # 使用isinstance替代type判断
            torch.nn.init.xavier_uniform_(m.weight)
    model.apply(init_weights)


    # 将模型移至指定设备
    model.to(device)
    model.train()

    # 定义优化器和损失函数
    optimizer = torch.optim.SGD(model.parameters(), lr=lr, momentum=0.9)
    criterion  = torch.nn.CrossEntropyLoss()

    best_acc = 0.0
    history = {'loss': [], 'val_loss': [], 'acc': [], 'val_acc': []}

    for epoch in range(epochs):
        model.train(True)
        running_loss = 0.0
        correct = 0
        total = 0


        for images, labels in train_loader:
            images, labels = images.to(device), labels.to(device)
            optimizer.zero_grad()
            outputs = model(images)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()

            running_loss += loss.item() * images.size(0)
            _, predicted = torch.max(outputs, 1)
            total += labels.size(0)
            correct += (predicted == labels).sum().item()

        train_acc = 100 * correct / total
        val_loss, val_acc = evaluate_accuracy(model, valid_loader, criterion, device)
        history['loss'].append(running_loss / total)
        history['val_loss'].append(val_loss)
        history['acc'].append(train_acc)
        history['val_acc'].append(val_acc)

        # 打印训练和验证的统计信息
        print(
            f'Epoch {epoch + 1}/{epochs}, Loss: {running_loss / total:.4f}, Train Acc: {train_acc:.2f}%, Val Loss: {val_loss:.4f}, Val Acc: {val_acc:.2f}%')

        # 如果当前验证准确率是最佳的，则保存模型
        if val_acc > best_acc:
            best_acc = val_acc
            save_best_model(model, optimizer, model_path)
            print(f"Model improved and saved to {model_path}")



"""
3.try_gpu(i=0) 
    如果存在，则返回指定的GPU设备，否则返回CPU设备。  
  
    Args:  
        i (int, optional): GPU设备的索引。默认为0。  
  
    Returns:  
        torch.device: 指定的GPU设备或CPU设备。 

"""
def try_gpu(i=0):
    # @save 注解：这个通常用于Jupyter Notebook中，用于保存函数定义到某个地方，但在这里它对代码执行没有影响。
    """如果存在，则返回gpu(i)，否则返回cpu()"""
    # 检查是否有足够的GPU设备数量，至少要比指定的索引i大1
    if torch.cuda.device_count() >= i + 1:
        # 如果有足够的GPU设备，则返回指定索引i的GPU设备
        return torch.device(f'cuda:{i}')
        # 如果没有足够的GPU设备或没有检测到GPU，则返回CPU设备
    return torch.device('cpu')










"""  
4.try_all_gpus()
   返回所有可用的GPU设备列表，如果没有GPU，则返回包含CPU设备的列表。  

   Returns:  
       list[torch.device]: 所有GPU设备或CPU设备的列表。  
"""

def try_all_gpus():
    # @save 注解：同样地，这个注解通常用于Jupyter Notebook中，用于保存函数定义。
    """返回所有可用的GPU，如果没有GPU，则返回[cpu(),]"""
    # 初始化一个空列表用于存储所有GPU设备
    devices = []
    # 遍历所有检测到的GPU设备
    for i in range(torch.cuda.device_count()):
        # 将每个GPU设备添加到列表中
        devices.append(torch.device(f'cuda:{i}'))
        # 如果devices列表不为空（即存在GPU设备），则返回该列表
    # 否则，返回只包含CPU设备的列表
    return devices if devices else [torch.device('cpu')]


















from ..environment import torch









"""
1.grad_clipping(net, theta)
定义一个函数来裁剪模型的梯度， 模型是从零开始实现的模型或由高级API构建的模型。 我们在此计算了所有模型参数的梯度的范数。
"""
def grad_clipping(net, theta):
    """裁剪梯度"""
    if isinstance(net, torch.nn.Module):
        params = [p for p in net.parameters() if p.requires_grad]
    else:
        params = net.params

    norm = torch.sqrt(sum(torch.sum((p.grad ** 2)) for p in params))
    if norm > theta:
        for param in params:
            param.grad[:] *= theta / norm


# callbacks 模块
# 功能：包含训练过程中的回调函数，用于实现如学习率调整、模型保存等功能。
# 子模块/文件：
# learning_rate_scheduler.py：学习率调整策略的实现。
# custom_callback.py：用户自定义回调函数的示例或模板。














# callbacks/__init__.py

# 导入 callback 子包中的各个模块
# from . import logging_callback
# from . import model_checkpoint_callback
# from . import early_stopping_callback
#
# 如果需要，可以直接导入这些模块中的特定回调类
from .custom_callback import save_best_model
from .custom_callback import load_best_model
# from .logging_callback import LoggingCallback
# from .model_checkpoint_callback import ModelCheckpointCallback
# from .early_stopping_callback import EarlyStoppingCallback
#
# # __all__ 变量定义了当使用 from callback import * 时导入哪些对象
# # 注意：通常不推荐使用 from package import *
__all__ = [
    'save_best_model',
    'load_best_model'
]


# models 模块
# 功能：包含各种深度学习模型的实现。
# 子模块/文件：
# cnn.py：卷积神经网络（CNN）相关模型的实现。
# rnn.py：循环神经网络（RNN）相关模型的实现。
# transformer.py：Transformer模型及其变体的实现。
# custom_model.py：用户自定义模型的示例或模板。

# models/__init__.py
# 如果需要，您可以直接导入这些模块中的特定类或函数
from .cnn import LeNet5
from .cnn import AlexNet
from .cnn import VGGBlock
from .cnn import VGG16
from .cnn import MLPConv
from .cnn import NiNBlock
from .cnn import NiN
from .cnn import InceptionBlockV1
from .cnn import GoogLeNet
from .cnn import ResidualBlock
from .cnn import ResNet
from .cnn import DenseBlock
from .cnn import TransitionBlock
from .cnn import DenseNet
from .cnn import YOLOv3_104
from .rnn import RNNModel







#from .rnn import RNNModel
#from .transformer import TransformerModel
#from .custom_model import CustomModel


# __all__ 变量定义了当使用 from models import * 时导入哪些对象
# 注意：通常不推荐使用 from package import *
__all__ = [
   'LeNet5',
   'AlexNet',
   'VGGBlock',
   'VGG16',
   'MLPConv',
   'NiNBlock',
   'NiN',
   'InceptionBlockV1',
   'GoogLeNet',
   'ResidualBlock',
   'ResNet',
   'DenseBlock',
   'TransitionBlock',
   'DenseNet',
   'YOLOv3_104',




   'RNNModel'
]









import setuptools

requirements = [
    'jupyter>=1.0.0',
    'numpy>=1.21.5',
    'matplotlib>=3.5.1',
    'requests>=2.25.1',
    'pandas>=1.2.4',
    'torch',
    'torchvision',
    'opencv-python',
]

setuptools.setup(
    name="PepperPepper",
    version="0.0.7.25",
    python_requires='>=3.11',
    author="Aohua Li",
    author_email="pepper2024jlu@163.com",
    description="It is a DeepLearning package to foster developed by Aohua Li",
    url="",
    packages=setuptools.find_packages(),
    zip_safe=True,
    install_requires=requirements,
)


